"""IFileOperation is fully wrapped in pywin32, so this is purely for typehinting."""

__all__ = [
    'IFileOperation',
]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Generic, TypeAlias, TypeVar

    from win32com.shell import shell  # type: ignore (module *does* exist)

    class IUnknown:
        def AddRef(self) -> int:
            ...

        def QueryInterface(self, pp) -> int:
            ...

        def Release(self) -> int:
            ...

    TIShellItem = TypeVar('TIShellItem')

    class _IFileOperation(Generic[TIShellItem], IUnknown):
        def Advise(self, pfops) -> int:
            ...

        def ApplyPropertiesToItem(self, item: TIShellItem) -> int:
            ...

        def ApplyPropertiesToItems(self, items: list[TIShellItem]) -> int:
            ...

        def CopyItem(
            self,
            item: TIShellItem,
            destination_folder: TIShellItem,
            new_name: str | None = ...,
        ) -> int:
            ...

        def CopyItems(
            self, items: list[TIShellItem], destination_folder: TIShellItem
        ) -> int:
            ...

        def DeleteItem(self, item: TIShellItem, pfofsItem=...) -> int:
            ...

        def DeleteItems(self, items: list[TIShellItem]) -> int:
            ...

        def GetAnyOperationsAborted(self, pfAnyOperationsAborted=...) -> int:
            ...

        def MoveItem(
            self,
            item: TIShellItem,
            destination_folder: TIShellItem,
            new_name: str | None = ...,
            pfopsItem=...,
        ) -> int:
            ...

        def MoveItems(
            self, items: list[TIShellItem], destination_folder: TIShellItem
        ) -> int:
            ...

        def NewItem(
            self,
            destination_folder: TIShellItem,
            file_attributes: int,
            name: str,
            pszTemplateName: str,
            pszDestinationName: str,
            pfopsItem=...,
        ) -> int:
            ...

        def PerformOperations(self) -> int:
            ...

        def RenameItem(self, item: TIShellItem, new_name: str, pfsopsItem=...) -> int:
            ...

        def RenameItems(self, items: list[TIShellItem], new_name: str) -> int:
            ...

        def SetOperationFlags(self, operation_flags: int) -> int:
            ...

        def SetOwnerWindow(self, hwndOwner) -> int:
            ...

        def SetProgressDialog(self, popd) -> int:
            ...

        def SetProgressMessage(self, message: str) -> int:
            ...

        def SetProperties(self, pPropArray) -> int:
            ...

        def Unadvise(self, dwCookie: int) -> int:
            ...

    def _make(
        callable: Callable[..., TIShellItem]
    ) -> type[_IFileOperation[TIShellItem]]:
        """Uses typing hijinks to extract the return type of a callable to apply to the
        IFileOperation interface.  This way typeshed hint types can be extracted and
        used as a hint for the member method parameters.
        """
        return _IFileOperation[TIShellItem]

    IFileOperation: TypeAlias = _make(shell.SHCreateItemFromParsingName)  # type: ignore
else:
    from typing import NewType

    IFileOperation = NewType('IFileOperation', object)



__all__ = [
    'IFileOperationProgressSink',
]

from ctypes import c_long
from ctypes.wintypes import DWORD, LPCWSTR, UINT

from comtypes import COMMETHOD, GUID, HRESULT, IUnknown

from .ishellitem import PIShellItem


class IFileOperationProgressSink(IUnknown):
    """Define the IFileOperationSink interface using comtypes, see
    https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ifileoperationprogresssink
    https://github.com/tpn/winsdk-10/blob/9b69fd26ac0c7d0b83d378dba01080e93349c2ed/Include/10.0.16299.0/um/ShObjIdl_core.h#L9652
    """

    _iid_ = GUID('{04b0f1a7-9490-44bc-96e1-4296a31252e2}')
    _methods_ = [
        COMMETHOD([], HRESULT, 'StartOperations'),
        COMMETHOD([], HRESULT, 'FinishOperations', (['in'], c_long, 'hrResult')),
        COMMETHOD(
            [],
            HRESULT,
            'PreRenameItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PostRenameItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['string', 'in'], LPCWSTR, 'pszNewName'),
            (['in'], HRESULT, 'hrRename'),
            (['in'], PIShellItem, 'psiNewlyCreated'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PreMoveItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PostMoveItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
            (['in'], c_long, 'hrMove'),
            (['in'], PIShellItem, 'psiNewlyCreated'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PreCopyItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PostCopyItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
            (['in'], HRESULT, 'hrCopy'),
            (['in'], PIShellItem, 'psiNewlyCreated'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PreDeleteItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PostDeleteItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiItem'),
            (['in'], HRESULT, 'hrDelete'),
            (['in'], PIShellItem, 'psiNewlyCreated'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PreNewItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'PostNewItem',
            (['in'], DWORD, 'dwFlags'),
            (['in'], PIShellItem, 'psiDestinationFolder'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszNewName'),
            (['string', 'unique', 'in'], LPCWSTR, 'pszTemplateName'),
            (['in'], DWORD, 'dwFileAttributes'),
            (['in'], HRESULT, 'hrNew'),
            (['in'], PIShellItem, 'psiNewItem'),
        ),
        COMMETHOD(
            [],
            HRESULT,
            'UpdateProgress',
            (['in'], UINT, 'iWorkTotal'),
            (['in'], UINT, 'iWorkSoFar'),
        ),
        COMMETHOD([], HRESULT, 'ResetTimer'),
        COMMETHOD([], HRESULT, 'PauseTimer'),
        COMMETHOD([], HRESULT, 'ResumeTimer'),
    ]



__all__ = [
    'convert_exceptions',
    'com_ptr',
]

import struct
from ctypes import POINTER, PyDLL, byref, c_bool, c_void_p, py_object
from functools import wraps
from typing import Callable, ParamSpec, TypeVar

import pythoncom
from comtypes import COMObject, IUnknown

from ..errors import (
    FileOperatorError,
    IFO_NotADirectoryError,
    InterfaceNotImplementedError,
    UserCancelledError,
)
from ..flags import FileOperationResult

PIUnknown = POINTER(IUnknown)  # type: ignore
P = ParamSpec('P')
T = TypeVar('T')


_PYCOM_PyObjectFromIUnknown = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown
_PYCOM_PyObjectFromIUnknown.restype = py_object
_PYCOM_PyObjectFromIUnknown.argtypes = (PIUnknown, c_void_p, c_bool)


def com_ptr(com_object: COMObject):
    return _PYCOM_PyObjectFromIUnknown(com_object, byref(IUnknown._iid_), True)


_hresult_to_exception = {
    FileOperationResult.E_DESTINATION_IS_FILE: NotADirectoryError,
    FileOperationResult.E_DRIVE_NOT_FOUND: IFO_NotADirectoryError,
    FileOperationResult.E_DESTINATION_IS_FOLDER: IsADirectoryError,
    FileOperationResult.E_ACCESS_DENIED_DESTINATION: PermissionError,
    FileOperationResult.E_ACCESS_DENIED_SOURCE: PermissionError,
    FileOperationResult.E_REQUIRES_ELEVATION: PermissionError,
    FileOperationResult.E_ALREADY_EXISTS_FOLDER: FileExistsError,
    FileOperationResult.E_ALREADY_EXISTS_NORMAL: FileExistsError,
    FileOperationResult.E_ALREADY_EXISTS_READONLY: FileExistsError,
    FileOperationResult.E_ALREADY_EXISTS_SYSTEM: FileExistsError,
    FileOperationResult.E_USER_CANCELLED: UserCancelledError,
    FileOperationResult.E_NOT_IMPLEMENTED: InterfaceNotImplementedError,
    # NOTE: FileNotFound handled by parse_name
}


def convert_exceptions(callable: Callable[P, T]) -> Callable[P, T]:
    """Wraps a method to automatically convert pythoncom.com_error exceptions
    into FileOperatorError exceptions.
    """

    # TODO: Implement mapping to builtin Python exceptions for common
    # HRESULTs, e.g. E_ACCESSDENIED, E_OUTOFMEMORY, etc.
    @wraps(callable)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return callable(*args, **kwargs)
        except pythoncom.com_error as e:
            # pywin32 defines it's HRESULTS using signed integers, but all the
            # definitions online use unsigned.
            hresult = e.hresult  # type: ignore
            if hresult < 0:
                hresult = struct.unpack('I', struct.pack('i', hresult))[0]
            if hresult in _hresult_to_exception:
                raise _hresult_to_exception[hresult](*e.args) from e
            else:
                raise FileOperatorError(*e.args) from e

    return wrapped



__all__ = [
    'IShellItem',
    'PIShellItem',
]

from ctypes import POINTER
from ctypes.wintypes import DWORD, LPWSTR

from comtypes import COMMETHOD, GUID, HRESULT, IUnknown


class _IShellItem(IUnknown):
    """partially define the IShellItem interface using comtypes, only GetDisplayName
    is implemented, see
    https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ishellitem
    https://github.com/tpn/winsdk-10/blob/9b69fd26ac0c7d0b83d378dba01080e93349c2ed/Include/10.0.16299.0/um/ShObjIdl_core.h#L7933
    """

    _iid_ = GUID('{43826d1e-e718-42ee-bc55-a1e261c37bfe}')
    _methods_ = [
        COMMETHOD([], HRESULT, 'BindToHandler'),
        COMMETHOD([], HRESULT, 'GetParent'),
        COMMETHOD(
            [],
            HRESULT,
            'GetDisplayName',
            (['in'], DWORD, 'sigdnName'),
            (['annotation', 'string', 'out'], POINTER(LPWSTR), 'ppszName'),
        ),
        COMMETHOD([], HRESULT, 'GetAttributes'),
    ]


PIShellItem = POINTER(_IShellItem)  # type: ignore


class IShellItem:
    """IShellItem interface, for typehints only"""

    def GetDisplayName(self, sigdnName: int) -> str:
        ...



__all__ = [
    'FOLDER_BIND_CTX',
    'parse_filename',
]
from ctypes import byref
from ctypes.wintypes import DWORD, WIN32_FIND_DATAW
from os import PathLike, fspath
from typing import TypeAlias

import pythoncom
import pywintypes
from comtypes import COMObject
from comtypes.hresult import S_OK
from win32com.shell import shell  # type: ignore (module *does* exist)

from ..flags import FileAttributeFlags
from .common import com_ptr
from .interfaces import IFileSysBindData

StrPath: TypeAlias = str | PathLike[str]


class FileSysBindData(COMObject):
    """Implement the IFileSysBindData interface"""

    _com_interfaces_ = [IFileSysBindData]

    def SetFindData(self, pfd) -> int:
        self.find_data = pfd
        return S_OK

    def GetFindData(self, ppfd) -> int:
        ppfd.contents = self.find_data
        return S_OK


def create_bind_ctx(find_data: WIN32_FIND_DATAW):
    """Create an IBindCtx instance with a custom WIN32_FIND_DATA structure."""
    # NOTE: Need the type-hint because of COMObject's __new__ method return types
    bind_data: FileSysBindData = FileSysBindData()
    bind_data.SetFindData(byref(find_data))
    bind_ctx = pythoncom.CreateBindCtx()
    bind_ctx.RegisterObjectParam('File System Bind Data', com_ptr(bind_data))
    return bind_ctx


FOLDER_BIND_CTX = create_bind_ctx(WIN32_FIND_DATAW(DWORD(FileAttributeFlags.DIRECTORY)))


E_FILE_NOT_FOUND = -2147024894


def parse_filename(path: StrPath, force: bool = False):
    """Parse a filename into an IShellItem2 instance. If `force` is True, then
    an IShellItem2 instance will be returned even if the file does not exist.
    """
    path = fspath(path)
    try:
        return shell.SHCreateItemFromParsingName(
            path, None, shell.IID_IShellItem2  # type: ignore
        )
    except pywintypes.com_error as e:
        if e.hresult == E_FILE_NOT_FOUND:  # type: ignore
            if force:
                return shell.SHCreateItemFromParsingName(
                    path, FOLDER_BIND_CTX, shell.IID_IShellItem2
                )
            else:
                raise FileNotFoundError(path) from None
        else:
            raise
